#!/usr/bin/python

from nilmdb.utils.printf import *

import threading
import subprocess
import cStringIO
import sys
import os
import signal
import time
import uuid
import psutil
import tempfile
import atexit
import shutil

class ProcessError(Exception):
    pass

class LogReceiver(object):
    """Spawn a thread that listens to a pipe for log messages,
    and stores them locally."""
    def __init__(self, pipe):
        self.pipe = pipe
        self.log = cStringIO.StringIO()
        self.thread = threading.Thread(target = self.run)
        self.thread.start()

    def run(self):
        while True:
            data = os.read(self.pipe, 65536)
            if not data:
                os.close(self.pipe)
                return
            self.log.write(data)

    def getvalue(self):
        return self.log.getvalue()

    def clear(self):
        self.log = cStringIO.StringIO()

class Process(object):
    """Spawn and manage a subprocess, and capture its output."""
    def __init__(self, name, argv, tempfile = None):
        self.start_time = None
        self.name = name

        # Use a pipe for communicating log data
        (rpipe, wpipe) = os.pipe()
        self._log = LogReceiver(rpipe)

        # Stdin is null
        nullfd = os.open(os.devnull, os.O_RDONLY)

        # Spawn the new process
        try:
            self._process = subprocess.Popen(args = argv, stdin = nullfd,
                                             stdout = wpipe, stderr = wpipe,
                                             close_fds = True, cwd = "/tmp")
        except (OSError, TypeError) as e:
            raise ProcessError(str(e))
        finally:
            # Close the FDs we don't need
            os.close(wpipe)
            os.close(nullfd)

        # Get process info
        self.start_time = time.time()
        self.pid = str(uuid.uuid1(self._process.pid or 0))

    def _join(self, timeout = 1.0):
        start = time.time()
        while True:
            if self._process.poll() is not None:
                return True
            if (time.time() - start) >= timeout:
                return False
            time.sleep(0.1)

    def terminate(self, timeout = 1.0):
        """Terminate a process, and all of its children that are in the same
        process group."""
        # First give it some time to die on its own
        if self._join(timeout):
            return True

        def getpgid(pid):
            try:
                return os.getpgid(pid)
            except OSError: # pragma: no cover
                return None

        # Find all children
        group = getpgid(self._process.pid)
        main = psutil.Process(self._process.pid)
        allproc = [ main ] + main.get_children(recursive = True)

        # Kill with SIGTERM, if they're still in this process group
        for proc in allproc:
            if getpgid(proc.pid) == group:
                os.kill(proc.pid, signal.SIGTERM)

        # Wait for it to die again
        if self._join(timeout):
            return True

        # One more try with SIGKILL
        for proc in allproc:
            if getpgid(proc.pid) == group:
                os.kill(proc.pid, signal.SIGKILL)

        # See if it worked
        return self._join(timeout)

    def clear_log(self):
        self._log.clear()

    @property
    def log(self):
        return self._log.getvalue()

    @property
    def alive(self):
        return self._process.poll() is None

    @property
    def exitcode(self):
        return self._process.returncode

class ProcessManager(object):
    """Track and manage a collection of Process objects"""
    def __init__(self):
        self.processes = {}
        self.tmpfiles = {}
        self.tmpdir = tempfile.mkdtemp(prefix = "nilmrun-usercode-")
        atexit.register(shutil.rmtree, self.tmpdir)

    def __iter__(self):
        return iter(self.processes.keys())

    def __getitem__(self, key):
        return self.processes[key]

    def run_code(self, procname, code, args):
        """Evaluate 'code' as if it were placed into a Python file and
        executed.  The arguments, which must be strings, will be
        accessible in the code as sys.argv[1:]."""
        # The easiest way to do this, by far, is to just write the
        # code to a file.
        (fd, path) = tempfile.mkstemp(prefix = "nilmrun-usercode-",
                                      suffix = ".py", dir=self.tmpdir)
        with os.fdopen(fd, 'w') as f:
            f.write(code)
        argv = [ sys.executable, "-B", "-s", "-u", path ] + args
        pid = self.run_command(procname, argv)
        self.tmpfiles[pid] = path
        return pid

    def run_command(self, procname, argv):
        """Execute a command line program"""
        new = Process(procname, argv)
        self.processes[new.pid] = new
        return new.pid

    def terminate(self, pid):
        return self.processes[pid].terminate()

    def remove(self, pid):
        if pid in self.tmpfiles:
            try:
                os.unlink(self.tmpfiles[pid])
            except OSError: # pragma: no cover
                pass
            del self.tmpfiles[pid]
        del self.processes[pid]
